package cam.codegen;
import java.io.*;
import java.lang.reflect.*;
import java.util.Vector;
import java.util.StringTokenizer;
/**
This class provides a programming interface to the Java class
reflection methods. The class to be interrogated is specifed
with the <a href = #setClass">setClass(...)</a> method.
@author Chris Anderson (C) UCLA 1998
@version 9/18/98
*/
public class ClassInterrogator
{
public ClassInterrogator()
{
    theClass    = null;
    packageName = null;
}
/**
Sets the class to be interrogated. The class name must be the full class name
(e.g. the name of the class including the package).
The class must reside in a directory that can be
located with the CLASSPATH/"package name" construct.
@exception java.lang.ClassNotFoundException if class is not found.
*/
public void setClass(String Cname) throws ClassNotFoundException
{
    className = new String(Cname);
    extractClassInfo();
}
/**
Returns the class name (without the package preface)
*/
public String getClassName()
{
    return localClassName;
}

/**
Returns the package name.
*/
public String getPackageName()
{
    return packageName;
}

/**
Returns the super class (parent class) name.
*/
public String getSuperclassName()
{
    return superClass;
}

/**
Returns the interfaces that the class implements.
*/
public String [] getInterfaceStrings()
{
    return interfaceStrings;
}
/**
Returns the number of public methods that the class
implements. Static methods, and constructors are excluded.
*/
public int getMethodCount()
{
    return methodCount;
}
/**
Returns the public method names.
Static methods and constructors are excluded.
*/
public String [] getMethodNames()
{
    return methodString;
}
/**
Returns the names of the objects that are
returned by the public methods.
*/
public String [] getReturnNames()
{
    return returnString;
}
/**
Returns the exceptions that are thrown by the public method.
@param methodNumber The index of the method. The method index corresponds to the
index of the method in the array of method names obtained with
getMethodNames().
*/
public String [] getExceptionNames(int methodNumber)
{
    return exceptionStrings[methodNumber];
}
/**
Returns the parameters of the method.
@param methodNumber The index of the method. The method index corresponds to the
index of the method in the array of method names obtained with
getMethodNames().
*/
public String [] getParameterNames(int methodNumber)
{
    return parameterStrings[methodNumber];
}
/**
Returns default argument names for the method parameters.
@param methodNumber The index of the method. The method index corresponds to the
index of the method in the array of method names obtained with
getMethodNames().
*/
public String [] getArgumentNames(int methodNumber)
{
    return argumentStrings[methodNumber];
}


private void extractClassInfo() throws ClassNotFoundException
{
     int i; int j;
     int mod;
//
//   Load the class. It needs to be in a directory within the classpath specification
//
//
     theClass = null;
     try{theClass = Class.forName(className);}
     catch(Exception e){System.out.println("Error " + e);};
//
//   Create a "local" name by stripping the package designation from the class
//
     String sTmp;
     StringTokenizer Tok;
     Tok  = new StringTokenizer(new String(className),".");
	 sTmp  = Tok.nextToken();
	 while(Tok.hasMoreTokens()){sTmp = Tok.nextToken();}
	 localClassName   = new String(sTmp);

	 Tok  = new StringTokenizer(new String(className),".");
	 int tokCount  = Tok.countTokens();
	 packageName     = null;
	 StringBuffer pB = new StringBuffer();
	 if(tokCount != 1)
	 {
	    pB.append(Tok.nextToken());
	    for(i = 2; i < tokCount; i++) {pB.append("." + Tok.nextToken());}
	    packageName = pB.toString();
	 }
//
//   Determine the superclass
//
     Class SClass = (theClass.getSuperclass());
     superClass = null;
     if(SClass != null)
     {
        superClass = new String(SClass.getName());
        if(superClass.equals("java.lang.Object")) superClass = null;
     }
//
//   Determine the interfaces that the interface class will implement
//
     interfaceStrings = null;
     Class [] interfaceClasses  = theClass.getInterfaces();
     if(interfaceClasses.length > 0)
     {
     interfaceStrings = new String[interfaceClasses.length];
     for(i = 0; i < interfaceClasses.length; i++)
     {interfaceStrings[i] = getTypeName(interfaceClasses[i]);}
     }
     //
     // Scan the methods of the class and select only public, non-static,
     // members for the the interface
     //
     Method [] methodTmp   = theClass.getDeclaredMethods();
     Vector methodVector = new Vector();
     for(i = 0; i < methodTmp.length; i++)
     {
        mod = methodTmp[i].getModifiers();
	    if(mod == 0) methodVector.addElement(methodTmp[i]);
	    else
	    if((Modifier.isPublic(mod))&&(!Modifier.isStatic(mod)))methodVector.addElement(methodTmp[i]);
	 }

	 Method method = null;
     methodCount   = methodVector.size();
     methodArray   = new Method[methodCount];
     for(i = 0; i < methodCount; i++)
     {
        methodArray[i] = (Method)methodVector.elementAt(i);
	 }

     methodString     = new String [methodCount];
     returnString     = new String [methodCount];
     exceptionStrings = new String [methodCount][];
     parameterStrings = new String [methodCount][];
     argumentStrings  = new String [methodCount][];

//
//   Access the methods individually and create string representations of
//   their components.
//
     int jp1;
     for(i = 0; i < methodCount; i++)
     {
        method = methodArray[i];

	    returnString[i]   = new String(getTypeName(method.getReturnType()));
	    methodString[i]   = new String(method.getName());
	    //
	    // parameters
	    //
	    Class[] params = method.getParameterTypes();
	    parameterStrings[i] = new String[params.length];
	    argumentStrings[i]  = new String[params.length];
	    for (j = 0; j < params.length; j++)
	    {
	    parameterStrings[i][j] = getTypeName(params[j]);
	    jp1 = j + 1;
	    argumentStrings[i][j]  = new String("Arg_" + jp1);
	    }
	    //
	    // exceptions
	    //
	    Class[] excepts = method.getExceptionTypes();
	    exceptionStrings[i] = new String[excepts.length];
	    for (j = 0; j < excepts.length; j++)
	    {
	    exceptionStrings[i][j] = getTypeName(excepts[j]);
	    }
    }
}
//
// Routine to repackage the names of the classes
//
private String getTypeName(Class type)
{
if (type.isArray())
    {
	    try {
		Class cl = type;
		int dimensions = 0;
		while (cl.isArray()) {
		    dimensions++;
		    cl = cl.getComponentType();
		}
		StringBuffer sb = new StringBuffer();
		sb.append(cl.getName());
		for (int i = 0; i < dimensions; i++) {
		    sb.append("[]");
		}
		return sb.toString();
	    } catch (Throwable e){}
	}
	return type.getName();
}

String packageName;

Class  theClass;
String className;
String localClassName;  // name without package appendage

String superClass;;
String [] interfaceStrings;

int       methodCount;
Method [] methodArray;

String []   methodString;      // names of methods (size methodCount)
String []   returnString;      // names of return object (size methodCount);

String [][] exceptionStrings;  //  1st index = method index; 2nd index = exception names
String [][] parameterStrings;  //  1st index = method index; 2nd index = prameters names
String [][] argumentStrings;   //  1st index = method index; 2nd index = argument  names


static public void main(String args[])
{

    int i; int j;
    String inputClass;

    if(args.length == 0) return;
    inputClass = args[0];

    ClassInterrogator R = new ClassInterrogator();
    try{R.setClass(inputClass);}
    catch(ClassNotFoundException e){System.out.println(e.getMessage());}
//
//  Display information about the class
//
    System.out.println(R.getPackageName());
    System.out.println(R.getClassName());

    String superClass = R.getSuperclassName();
    if(superClass != null) System.out.println(superClass);

    String [] interfaceStrings = R.getInterfaceStrings();
    if(!(interfaceStrings == null))
    {
    for(i = 0; i < interfaceStrings.length; i++)
    {
        System.out.print(interfaceStrings[i] + "  " );
    }
    System.out.println(" ");
    }

    int methodCount = R.getMethodCount();
    System.out.println(" Method Count " + methodCount);

    String [] methodNames = R.getMethodNames();
    String [] returnNames = R.getReturnNames();
    String [] parameterNames;
    String [] argumentNames;
    String [] exceptionNames;
    for(i = 0; i < methodCount; i++)
    {
        System.out.print(returnNames[i] + "  ");
        System.out.print(methodNames[i] + "  ");
        parameterNames = R.getParameterNames(i);
        argumentNames  = R.getArgumentNames(i);
        exceptionNames = R.getExceptionNames(i);

        for(j = 0; j < parameterNames.length; j++)
        {
            System.out.print(parameterNames[j] + " ");
            System.out.print(argumentNames[j]  + " ");
        }
        for(j = 0; j < exceptionNames.length; j++)
        {
            System.out.print(exceptionNames[j] + " ");
        }
        System.out.println(" ");
    }


    System.out.println("Done : Hit return to continue ");
    BufferedReader D =
    new BufferedReader(new InputStreamReader(System.in));
    try{D.readLine();}catch(Exception e){}
}




}