package cam.codegen;
import java.io.*;
import java.util.Vector;
import java.util.Date;
import java.text.DateFormat;

//
//##################################################################
//  Class  CreateNetAppWrappers
//##################################################################
// Chris Anderson (C) UCLA 1998
// 9/22/98
/**
A class for creating remote and server wrappers to be used with
the cam.netapp distributed application infrastructure. See
<A HREF="../../NetappWrappers/NetAppWrap.html"target="_parent">NetApp Wrappers</A> 
for documentation about how to construct distributed applications using 
these wrapper classes.
<pre>
#   Input  <-- className
#    
#   Output --> classNameCI (client class implementation)
#          --> classNameSI (server class implementation)
</pre>

The creation of the wrapper classes must be re-done every time the
method interface to the input class changes.<p>

Target and derived class locations for use with cam.netapp infrastructure
<pre>
Local                                   |        Remote
-------                                 |      ----------
Client Application (uses classNameCI)   |       className.class
ClassNameCI.class                       |       classNameSI.class
</pre>
@author Chris Anderson (C) UCLA 1998
@version 9/22/98
*/
public class CreateNetAppWrappers
{

/**
Sets the target class. 
@param cName Target class name (full classname - e.g. with package prefix if any).
@exception java.lang.ClassNotFoundException if target class is not found. 
*/
public void setClass(String cName) throws ClassNotFoundException
{
    className = cName;
    classInfo = new ClassInterrogator();
    classInfo.setClass(className);
}
/**
Returns the name of the target class without the package prefix.
*/
public String getLocalClassName()
{
    return classInfo.getClassName();
}
//
//#####################################################################
//            Create client implementation 
//#####################################################################
//
/**
Creates the client implementation class. 
@return a String array, each element being a line of the class code.
*/
public String [] createClientImplementation()
{

    int i; int j;
    StringBuffer sb;
    Vector cLines = new Vector();
    //
    // create package name
    //
    String packageName = classInfo.getPackageName();
    if(packageName != null)
    {cLines.addElement("package " + classInfo.getPackageName() + ";");}

    cLines.addElement("//");
    cLines.addElement("//#################################################################");
    cLines.addElement("//      Class " + classInfo.getClassName() +"CI  : " + classInfo.getClassName() + " (C)lient (I)mplementation ");
    cLines.addElement("//#################################################################");
    cLines.addElement("//");
    Date JobDate;
    JobDate = new Date();
    String dateOut =
    DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.SHORT).format(JobDate);
    cLines.addElement("// Construction Date : " + dateOut);
    cLines.addElement("// Auto-generated by CreateNetAppWrappers (Chris Anderson (C) UCLA 1998)");
    cLines.addElement("//");
    cLines.addElement("import java.io.*;");
    cLines.addElement("import java.net.Socket;");
    cLines.addElement("import java.util.Vector;");
    cLines.addElement("import cam.netapp.*;");
    cLines.addElement(" ");
    //
    //  class declaration : name
    //
    sb = new StringBuffer();
    sb.append("public class " + classInfo.getClassName() + "CI");
    //
    //  class declaration : extends (if it does)
    //
    String superClass = classInfo.getSuperclassName();
    if(superClass != null) sb.append(" extends " + superClass);
    //
    //  class declaration : implements (if it does)
    //
    String [] interfaceStrings = classInfo.getInterfaceStrings();
    if(!(interfaceStrings == null))
    {
    sb.append(" implements ");
    {sb.append(interfaceStrings[0]);}
    for(i = 1; i < interfaceStrings.length; i++)
    {sb.append(", " + interfaceStrings[i]);}
    sb.append(", cam.netapp.CIinterface");
    }
    else
    {
    sb.append(" implements cam.netapp.CIinterface");   
    }
    cLines.addElement(sb.toString());
    //
    //  body of class
    //
    cLines.addElement("{");
    cLines.addElement(" ");
    cLines.addElement("//");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//     Client Implementation of Source Class Methods  ");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//");
    //
    // capture names
    //
    String [] methodNames = classInfo.getMethodNames();
    String [] returnNames = classInfo.getReturnNames();
    String [] parameterNames;
    String [] argumentNames;
    String [] exceptionNames;
    int       exceptionCount;
    int       parameterCount;
    int       parameterCountp1;
    //
    // loop over all public methods, and create code for them
    //
    for(i = 0; i < classInfo.getMethodCount(); i++)
    {
       // compose method name :

       sb = new StringBuffer();

       // pubilc + return type  + method

       sb.append("public " + returnNames[i] + " " + methodNames[i] + "(");
       parameterNames = classInfo.getParameterNames(i);
       argumentNames  = classInfo.getArgumentNames(i);
       exceptionNames = classInfo.getExceptionNames(i);

       // ( parameters )

       parameterCount = parameterNames.length;
       if(parameterNames.length >= 1)
       {sb.append(parameterNames[0] + " " + argumentNames[0]);}
       for(j = 1; j < parameterNames.length; j++)
       {sb.append("," + parameterNames[j] + " " + argumentNames[j]);}
       sb.append(")");

       // exceptions

       exceptionCount = exceptionNames.length;
       if(exceptionCount >= 1)
       {sb.append(" throws " + exceptionNames[0]);}
        for(j = 1; j < exceptionCount; j++)
       {sb.append(", " + exceptionNames[j]);}


       cLines.addElement(sb.toString());
//
//     compose body
//
       cLines.addElement("{");
       
//
//     if not a setClient method, then compose a standard body
//
       if(!(sb.toString().equals("public void setClient(java.lang.Object Arg_1) throws java.lang.Exception")))
       {
        
       //  first declare Vector to pass the parameters

       parameterCountp1= parameterCount +1;
       cLines.addElement("    Vector methodData = new Vector(" +  parameterCountp1  + ");");

       // set method index as first object in parameter vector

       cLines.addElement("    methodData.addElement(new Integer(" + i + "));");

       // set parameters to be elements of the vector, converting primitive data types
       // into objects as necessary

       for(j = 0; j < parameterNames.length; j++)
       {
        if(isPrimitiveType(parameterNames[j]))
        {
        cLines.addElement("    methodData.addElement(new "
        + getClassForType(parameterNames[j]) +
        "(" + argumentNames[j] +"));");
        }
        else
        {cLines.addElement("    methodData.addElement(" + argumentNames[j] +");");}
       }
       // write method data to the remote routine

       cLines.addElement(" ");
       cLines.addElement("    try{Os.writeObject(methodData);}");
       cLines.addElement("    catch(Exception e){System.out.println(e);};");
       cLines.addElement(" ");

       // read return data

       cLines.addElement("    Vector MethodRet = new Vector();");
       cLines.addElement("    try{MethodRet = (Vector)Is.readObject();}");
       cLines.addElement("    catch(Exception e){System.out.println(e);};");
       cLines.addElement(" ");
       cLines.addElement("    int retValue = ((Integer)MethodRet.elementAt(0)).intValue();");
       cLines.addElement("    if(retValue != 0)");
       cLines.addElement("    {");
       //
       // catch exceptions : indicated with a non-zero return value. Return value -1 gives
       // the index of the exception to throw
       //
       if(exceptionCount >= 1)
       {
        cLines.addElement("      switch(retValue-1) ");
        cLines.addElement("      {");
        for(j = 0; j < exceptionCount; j++)
        {
        cLines.addElement("      case "+ j + " : " );
        cLines.addElement("      throw (" + exceptionNames[j] + ")MethodRet.elementAt(1);");
       }
        cLines.addElement("      }");
       }
       else
       {cLines.addElement("    System.out.println(\"Error In Remote Invocation Of " + classInfo.getClassName() + "."
       + methodNames[i] + "\");");}
       cLines.addElement("    }");

       // normal return, so unpack arguments and return

       if(returnNames[i].equals("void"))
       {
       cLines.addElement("    return;");
       }
       else if(isPrimitiveType(returnNames[i]))
       {
       cLines.addElement("    return (("+ getClassForType(returnNames[i])
       +")MethodRet.elementAt(1))." +
       returnNames[i] + "Value();");
       }
       else
       {
       cLines.addElement("    return ("+ returnNames[i]+")MethodRet.elementAt(1);");
       }
       
       } // close of if(! createClient methods)
       else
       {
        cLines.addElement("    Class theClass = Arg_1.getClass();");
        cLines.addElement("    String cName   = theClass.getName();");
        cLines.addElement("    Object theObject = null;");
        
        cLines.addElement("    ");
        
        cLines.addElement("    try");
        cLines.addElement("    {");
        cLines.addElement("    theClass   = Class.forName(cName + \"SI\");");
        cLines.addElement("    theObject  = theClass.newInstance();"); 
        cLines.addElement("    }");
        cLines.addElement("    catch(Exception ex)");
        cLines.addElement("    {throw new Exception(\"Server Class \" + cName + \"SI Not Found : \");};");
        cLines.addElement("    ");
        cLines.addElement("    ((SIinterface)theObject).setSIobject(Arg_1);");
        cLines.addElement("    ");
     
       cLines.addElement("    Socket [] auxSocket = new Socket[1];");
       cLines.addElement("    auxSocket[0]        = netConnect.getAuxSocket(1);");
       cLines.addElement("    NetworkConnection  auxConnection = new NetworkConnection(auxSocket);");
       cLines.addElement("    ((SetNetworkConnection)theObject).setNetworkConnection(auxConnection);");
        
        cLines.addElement("    ");
        cLines.addElement("    Thread SIrun = new Thread((Runnable)theObject);");
        cLines.addElement("    SIrun.start();");
        cLines.addElement("    ");
        cLines.addElement("    Vector methodData = new Vector(2);");
       // set method index as first object in parameter vector
        cLines.addElement("    methodData.addElement(new Integer(" + i + "));");
        cLines.addElement("    methodData.addElement(new String(cName + \"CI\"));");
        
       // write method data to the remote routine

       cLines.addElement(" ");
       cLines.addElement("    try{Os.writeObject(methodData);}");
       cLines.addElement("    catch(Exception e){System.out.println(e);};");
       cLines.addElement(" ");

       // read return data

       cLines.addElement("    Vector MethodRet = new Vector();");
       cLines.addElement("    try{MethodRet = (Vector)Is.readObject();}");
       cLines.addElement("    catch(Exception e){System.out.println(e);};");
       cLines.addElement(" ");
       cLines.addElement("    int retValue = ((Integer)MethodRet.elementAt(0)).intValue();");
       cLines.addElement("    if(retValue != 0)");
       cLines.addElement("    {");
       //
       // catch exceptions : indicated with a non-zero return value. Return value -1 gives
       // the index of the exception to throw
       //
       if(exceptionCount >= 1)
       {
        cLines.addElement("      switch(retValue-1) ");
        cLines.addElement("      {");
        for(j = 0; j < exceptionCount; j++)
        {
        cLines.addElement("      case "+ j + " : " );
        cLines.addElement("      throw (" + exceptionNames[j] + ")MethodRet.elementAt(1);");
       }
        cLines.addElement("      }");
       }
       else
       {cLines.addElement("    System.out.println(\"Error In Remote Invocation Of " + classInfo.getClassName() + "."
       + methodNames[i] + "\");");}
       cLines.addElement("    }");


        cLines.addElement("    return;");
       }
       // close of method creation loop
      
       cLines.addElement(" ");
       cLines.addElement("}");
       cLines.addElement(" ");
    }
    cLines.addElement("//");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//      Client Implementation Specific Methods ");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//");

    // constructor

    cLines.addElement("public " + classInfo.getClassName() + "CI()");
    cLines.addElement("{");
    cLines.addElement("    verboseFlag = false;");
    cLines.addElement("    Os          = null;");
    cLines.addElement("    Is          = null;");
    cLines.addElement("    netConnect = new NetworkConnection();");
    cLines.addElement("    netConnect.setVerboseFlag(verboseFlag);");
    cLines.addElement("}");
    //
    // finalize --- need to call disconnectRemoteApplication so that
    // RI class stops
    //

    cLines.addElement("protected void finalize() throws Throwable");
    cLines.addElement("{");
    cLines.addElement("    netConnect.disconnectRemoteApplication();");
    cLines.addElement("}");

    // setObjectStreams
    
    cLines.addElement("public void setObjectStreams(InputStream inS, OutputStream outS) throws Exception");
    cLines.addElement("{");
    cLines.addElement(" ");
    cLines.addElement("    try{Os = new ObjectOutputStream(outS);}");
    cLines.addElement("    catch(Exception e){throw e;}");
    cLines.addElement(" ");
    cLines.addElement("    try{Is = new ObjectInputStream(inS);}");
    cLines.addElement("    catch(Exception e){throw e;}");
    cLines.addElement(" ");
    cLines.addElement("}");

    
    // remote initialization routine

    cLines.addElement(" ");
    cLines.addElement("public void createServerInstance(String address, int portnum)");

    cLines.addElement("throws Exception");
    cLines.addElement("{");


    String serverClass = null;
    if(packageName != null)
    {
    serverClass = new String(classInfo.getPackageName() + "." + classInfo.getClassName() + "SI");
    }
    else
    {
    serverClass = new String(classInfo.getClassName() + "SI");
    }

    cLines.addElement("    int auxConnections = 1;");
    cLines.addElement("    try{netConnect.connectRemoteApplication(address, portnum,\"" +
                            serverClass + "\""+ ", auxConnections);}");
    cLines.addElement("    catch(Exception e)");
    cLines.addElement("    {throw new Exception(\"" + classInfo.getClassName() + "CI : Unable to Connect to Remote Instance\");};");
    cLines.addElement("    netConnect.startRemoteApplication();");
    cLines.addElement(" ");
    cLines.addElement("    try{Os = new ObjectOutputStream(netConnect.getOutputStream());}");
    cLines.addElement("    catch(Exception e)");
    cLines.addElement("    {throw new Exception(\"" + classInfo.getClassName() + "CI : Unable to Connect to Remote Instance\");};");
    cLines.addElement(" ");
    cLines.addElement("    try{Is = new ObjectInputStream(netConnect.getInputStream());}");
    cLines.addElement("    catch(Exception e)");
    cLines.addElement("    {throw new Exception(\"" + classInfo.getClassName() + "CI : Unable to Connect to Remote Instance\");};");
    cLines.addElement(" ");
    cLines.addElement("    if(verboseFlag){System.out.println(\"" + classInfo.getClassName() + "CI " + ": Connection Complete\");};");

    cLines.addElement("    ");
    cLines.addElement("}");

    // shutdown code

    cLines.addElement(" ");
    cLines.addElement("public void setVerboseFlag(boolean flag) ");
    cLines.addElement("{");
    cLines.addElement("    verboseFlag = flag;");
    cLines.addElement("    netConnect.setVerboseFlag(verboseFlag);");
    cLines.addElement("}");


    cLines.addElement(" ");
    //
    //  end of class body
    //
    cLines.addElement("//");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//      Data Members ");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//");
    cLines.addElement("    ObjectOutputStream Os;");
    cLines.addElement("    ObjectInputStream  Is;");
    cLines.addElement("    NetworkConnection netConnect;");
    cLines.addElement("    public boolean verboseFlag;");
    cLines.addElement("}");

    String [] returnString = new String[cLines.size()];
    for(i = 0; i < returnString.length; i++)
    {
    returnString[i] = (String)cLines.elementAt(i);
    }
    return returnString;
}
//
//#####################################################################
//                  CREATE SERVER CODE
//#####################################################################
//
/**
Creates the server implementation class. 
@return a String array, each element being a line of the class code.
*/
public String [] createServerCode()
{

    int i; int j;
    StringBuffer sb;
    Vector cLines = new Vector();

    // package name

    String packageName = classInfo.getPackageName();
    if(packageName != null)
    {cLines.addElement("package " + classInfo.getPackageName() + ";");}

    cLines.addElement("//");
    cLines.addElement("//#################################################################");
    cLines.addElement("//      Class " + classInfo.getClassName() +"SI  : " + classInfo.getClassName() + " (S)erver (I)mplementation ");
    cLines.addElement("//#################################################################");
    cLines.addElement("//");
    Date JobDate;
    JobDate = new Date();
    String dateOut =
    DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.SHORT).format(JobDate);
    cLines.addElement("// Construction Date : " + dateOut);
    cLines.addElement("// Auto-generated by CreateNetAppWrappers (Chris Anderson (C) UCLA 1998)");
    cLines.addElement("//");

    // imports

    cLines.addElement("import java.io.*;");
    cLines.addElement("import java.net.Socket;");
    cLines.addElement("import java.util.Vector;");
    cLines.addElement("import cam.netapp.SetNetworkConnection;");
    cLines.addElement("import cam.netapp.NetworkConnection;");
    cLines.addElement("import cam.netapp.SIinterface;");
    cLines.addElement("import cam.netapp.CIinterface;");

    cLines.addElement(" ");

    // class declaration

    sb = new StringBuffer();
    sb.append("public class " + classInfo.getClassName() + "SI implements Runnable, SetNetworkConnection, SIinterface");
    cLines.addElement(sb.toString());
    cLines.addElement("{");
    cLines.addElement(" ");

    //  body of class

    cLines.addElement("//");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//     Server Implementation of Source Class Methods  ");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//");

    //  capture names

    String [] methodNames = classInfo.getMethodNames();
    String [] returnNames = classInfo.getReturnNames();
    String [] parameterNames;
    String [] argumentNames;
    String [] exceptionNames;
    int       exceptionCount;
    int       parameterCount;
    int       parameterCountp1;

    // run() ... handler for the method calls

    cLines.addElement(" ");
    cLines.addElement("public void run()");
    cLines.addElement("{");

    cLines.addElement(" ");


    cLines.addElement("    int     methodIndex    = -1;");
    cLines.addElement("    boolean comErrorFlag   = false;");

    // set up input and output object streams

    cLines.addElement(" ");
    cLines.addElement("    Is = null;");
    cLines.addElement("    try");
    cLines.addElement("    {Is = new ObjectInputStream(InputS);}");
    cLines.addElement("    catch(Exception e){/*System.out.println(e);*/ comErrorFlag = true;};");
    cLines.addElement(" ");
    cLines.addElement("    Os = null;");
    cLines.addElement("    try");
    cLines.addElement("    {Os = new ObjectOutputStream(OutputS);}");
    cLines.addElement("    catch(Exception e){/*System.out.println(e);*/ comErrorFlag = true;};");
    cLines.addElement(" ");

    cLines.addElement(" ");
    cLines.addElement("    if(cObj == null) cObj = new " + classInfo.getClassName() + "();");
    cLines.addElement(" ");

    // Main processing loop : Methods are accessed via the method index = value in
    // the first object of the method data vector
    //
    cLines.addElement("    while(!comErrorFlag)");
    cLines.addElement("    {");
    cLines.addElement(" ");

    // capture method data

    cLines.addElement("    Vector methodData = new Vector();");
    cLines.addElement("    try");
    cLines.addElement("    {methodData = (Vector)Is.readObject();}");
    cLines.addElement("    catch(Exception e){/*System.out.println(e);*/ comErrorFlag = true;}");
    cLines.addElement(" ");

    // process method

    cLines.addElement("    if(!comErrorFlag)");
    cLines.addElement("    {");

    cLines.addElement("    methodIndex =  ((Integer)methodData.elementAt(0)).intValue();");
    cLines.addElement("    Vector methodRet = new Vector(2);");
    cLines.addElement(" ");
    cLines.addElement("    switch(methodIndex) {");

    int jp1 = 0;
    for(i = 0; i < classInfo.getMethodCount(); i++)
    {

       // capture method information

       parameterNames = classInfo.getParameterNames(i);
       argumentNames  = classInfo.getArgumentNames(i);
       exceptionNames = classInfo.getExceptionNames(i);
       methodNames    = classInfo.getMethodNames();
       parameterCount = parameterNames.length;
       
 //////
 //
 //    Compose the method name; to be used to identify the setClient method. This code
 //    is a replica of that used in the createClientImplementation
 //
       sb = new StringBuffer();

       // public + return type  + method

       sb.append("public " + returnNames[i] + " " + methodNames[i] + "(");
       parameterNames = classInfo.getParameterNames(i);
       argumentNames  = classInfo.getArgumentNames(i);
       exceptionNames = classInfo.getExceptionNames(i);

       // ( parameters )

       parameterCount = parameterNames.length;
       if(parameterNames.length >= 1)
       {sb.append(parameterNames[0] + " " + argumentNames[0]);}
       for(j = 1; j < parameterNames.length; j++)
       {sb.append("," + parameterNames[j] + " " + argumentNames[j]);}
       sb.append(")");

       // exceptions

       exceptionCount = exceptionNames.length;
       if(exceptionCount >= 1)
       {sb.append(" throws " + exceptionNames[0]);}
        for(j = 1; j < exceptionCount; j++)
       {sb.append(", " + exceptionNames[j]);}
//
//
/////

       // work on ith method

       cLines.addElement("      case " + i + " :   // " + methodNames[i] + "(...)");
       //
       // if not setClient, then create standard body
       //
       if(!(sb.toString().equals("public void setClient(java.lang.Object Arg_1) throws java.lang.Exception")))
       {
       //
       // normal variable unpacking
       //
       for(j = 0; j < parameterCount; j++)
       {
       jp1= j+1;
       if(isPrimitiveType(parameterNames[j]))
        {
        cLines.addElement("        " + parameterNames[j] + " " + argumentNames[j]+i+" = (("
        + getClassForType(parameterNames[j]) +
        ")methodData.elementAt(" + jp1 +"))." + parameterNames[j] + "Value();");
        }
        else
        {cLines.addElement("        " + parameterNames[j] + " "+ argumentNames[j]+ i+
        " = (" + parameterNames[j] + ")methodData.elementAt(" + jp1 +");");}
       }
       }
       else
       {
        cLines.addElement("        String cName = (String)methodData.elementAt(1);");
       }
       
       // method  invocation
       
       exceptionCount    = exceptionNames.length;
       if(exceptionCount >= 1)
       {
        cLines.addElement("        try");
        cLines.addElement("        {");
       }
       
       if(!(sb.toString().equals("public void setClient(java.lang.Object Arg_1) throws java.lang.Exception")))
       {
       sb = new StringBuffer();
       if(!(returnNames[i].equals("void")))
       {
        sb.append("        "+ returnNames[i] + " R_"+i+"= cObj." + methodNames[i] + "(");
       }
       else
       {
        sb.append("        cObj."+  methodNames[i] + "(");
       }
       if(parameterCount > 0) sb.append(argumentNames[0]+ i);
       for(j = 1; j < parameterCount; j++) {sb.append(", " + argumentNames[j]+ i);}
       sb.append(");");
       cLines.addElement(sb.toString());
       }
       else
       {
       // create CI class
         cLines.addElement("        Class theClass;");
         cLines.addElement("        Object theObject = null;");
         cLines.addElement("    ");
         cLines.addElement("        theClass   = Class.forName(cName);");
         cLines.addElement("        theObject  = theClass.newInstance();"); 
         cLines.addElement("        Socket auxSocket =  netConnect.getAuxSocket(1);");
         cLines.addElement("        ((CIinterface)theObject).setObjectStreams(auxSocket.getInputStream(),auxSocket.getOutputStream());");
         cLines.addElement("        "); 
         cLines.addElement("        cObj.setClient(theObject);");
       //
       }
       if(exceptionCount >= 1)
       {
        cLines.addElement("        methodRet.addElement(new Integer(0));");
        cLines.addElement("        }");
        cLines.addElement("        catch(Exception e)");
        cLines.addElement("        {");

        for(j = 0; j < exceptionCount; j++)
        {
        jp1= j+1;
        cLines.addElement("        if(e instanceof " + exceptionNames[j]
        + "){methodRet.addElement(new Integer(" + jp1 + "));");
        cLines.addElement("        methodRet.addElement(e);}");
        }
        
        cLines.addElement("        }");
       }
       if(exceptionCount == 0)
       {cLines.addElement("        methodRet.addElement(new Integer(0));");}

       // pack return data

       if(!(returnNames[i].equals("void")))
       {

        if(isPrimitiveType(returnNames[i]))
        {
        cLines.addElement("        methodRet.addElement(new "
        + getClassForType(returnNames[i]) +
        "(R_"+i+"));");
        }
        else
        {cLines.addElement("        methodRet.addElement(R_"+i+");");}
       }
       cLines.addElement("      break;");
     }
    cLines.addElement("    } //  switch ");
    cLines.addElement("    ");
    cLines.addElement("    try{Os.writeObject(methodRet);}");
    cLines.addElement("    catch(Exception e){/*System.out.println(e);*/; comErrorFlag = true;};");
    cLines.addElement("    ");

    cLines.addElement("    } //  If !comErrorFlag ");
    cLines.addElement("   } //   While !comErrorFlag ");
    cLines.addElement(" ");
    cLines.addElement("}     //  run ");

    cLines.addElement("//");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//      Server Implementation Specific Methods ");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//");
    cLines.addElement("public " + classInfo.getClassName() + "SI()");
    cLines.addElement("{");
    cLines.addElement("    cObj        = null;");
    cLines.addElement("    Os          = null;");
    cLines.addElement("    Is          = null;");
    cLines.addElement("}");

    cLines.addElement("public boolean setNetworkConnection(NetworkConnection N)");
    cLines.addElement("{");
    cLines.addElement("    InputS      = N.getInputStream();");
    cLines.addElement("    OutputS     = N.getOutputStream();");
    cLines.addElement("    netConnect  = N;");
    cLines.addElement("    return true;");
    cLines.addElement("}");

    cLines.addElement("public void setSIobject(Object Ob)");
    cLines.addElement("{");
    cLines.addElement("    cObj     = (" + classInfo.getClassName() + ")Ob;");
    cLines.addElement("    return;");
    cLines.addElement("}");
    
    cLines.addElement(" ");
    //
    //  end of class body
    //
    cLines.addElement("//");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//      Data Members ");
    cLines.addElement("//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++");
    cLines.addElement("//");
    cLines.addElement("    "+ classInfo.getClassName() + "  cObj;");
    cLines.addElement("    ObjectOutputStream Os;");
    cLines.addElement("    ObjectInputStream  Is;");
    cLines.addElement("    NetworkConnection netConnect;");
    cLines.addElement("    OutputStream OutputS;");
    cLines.addElement("    InputStream  InputS;");
    cLines.addElement("}");

    String [] returnString = new String[cLines.size()];
    for(i = 0; i < returnString.length; i++)
    {
    returnString[i] = (String)cLines.elementAt(i);
    }
    return returnString;
}
//
//#####################################################################
//                  UTILITY ROUTINES
//#####################################################################
//
boolean isPrimitiveType(String S)
{
    if(S.equals("double"))   return true;
    if(S.equals("int"))      return true;
    if(S.equals("boolean"))  return true;
    if(S.equals("char"))     return true;
    if(S.equals("float"))    return true;
    if(S.equals("byte"))     return true;
    if(S.equals("long"))     return true;
    if(S.equals("short"))    return true;
    return false;
}

String getClassForType(String S)
{
    if(S.equals("double"))   return new String("Double");
    if(S.equals("int"))      return new String("Integer");
    if(S.equals("boolean"))  return new String("Boolean");
    if(S.equals("char"))     return new String("Character");
    if(S.equals("float"))    return new String("Float");
    if(S.equals("byte"))     return new String("Byte");
    if(S.equals("long"))     return new String("Long");
    if(S.equals("short"))    return new String("Short");
    return null;
}

    String className;
    ClassInterrogator classInfo;
    int   setClientIndex;
/**
Command line invocation routine. Creates the the client
and server implementations. Compiles the client and server classes. 
@param args[] Target class; specified with a full classname 
(e.g. including a package prefix if any).
*/
static public void main(String args[])
{
    int i; int j;
    String inputClass;

    if(args.length == 0) return;
    inputClass = args[0];

    CreateNetAppWrappers codeCreator = new CreateNetAppWrappers();
    try{codeCreator.setClass(inputClass);}
    catch(ClassNotFoundException e){System.out.println(e.getMessage());}
    String localClassName = codeCreator.getLocalClassName();

    System.out.println("Building Client Class : " + localClassName + "CI");

    String [] clientCode  = codeCreator.createClientImplementation();
    //
    //Screen Output
    //
    /*
    for(i = 0; i < clientCode.length; i++)
    {
        System.out.println(clientCode[i]);
    }
    */

    File outFileI;
    PrintWriter fileOutputI = null;

    outFileI          = new File(localClassName + "CI.java");
    try{fileOutputI   = new PrintWriter(new FileOutputStream(outFileI),true);}
    catch(IOException e){};

    for(i = 0; i < clientCode.length; i++)
    {
     fileOutputI.println(clientCode[i]);
    }
    System.out.println("Finished Client Class : " + localClassName + "CI.java");
    System.out.println("Building Server Class : " + localClassName + "SI");

    String [] serverCode  = codeCreator.createServerCode();
    //
    //Screen Output
    //
    /*
    for(i = 0; i < serverCode.length; i++)
    {
        System.out.println(serverCode[i]);
    }
    */

    File outFileS;
    PrintWriter fileOutputS = null;

    outFileS          = new File(localClassName + "SI.java");
    try{fileOutputS   = new PrintWriter(new FileOutputStream(outFileS),true);}
    catch(IOException e){};

    for(i = 0; i < serverCode.length; i++)
    {
     fileOutputS.println(serverCode[i]);
    }
    System.out.println("Finished Server Class : " + localClassName + "SI.java");
    //
    // Invoke the compiler 
    //
    sun.tools.javac.Main compiler = new sun.tools.javac.Main(System.out, "javac");
        
    System.out.println("Compiling : " + localClassName + "CI.java");
    String [] cArgs1 = {localClassName + "CI.java","-O"};
    compiler.compile(cArgs1);
    
    System.out.println("Compiling : " + localClassName + "SI.java");
    String [] cArgs2 = {localClassName + "SI.java","-O"};
    compiler.compile(cArgs2);

    System.out.println("Done : Hit return to continue ");
    BufferedReader D =
    new BufferedReader(new InputStreamReader(System.in));
    try{D.readLine();}catch(Exception e){}
}

}
