package cam.codegen;
import java.io.*;
import java.lang.reflect.*;
import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Date;
import java.text.DateFormat;
import cam.codegen.ClassInterrogator;
//
//##################################################################
//  Class CreateRMIWrappers
//##################################################################
// Chris Anderson (C) UCLA 1998
// 9/22/98
/**
A class for creating an RMI interface and RMI server wrapper classes. 
See <A HREF="../../NetappWrappers/NetAppWrap.html"target="_parent">NetApp Wrappers</A> 
for documentation about how to construct distributed applications using 
these wrapper classes.
<pre>
#   Input  <-- className
#    
#   Output --> classNameRI (an RMI interface)
#          --> classNameRS (an RMI server class)
</pre>

The creation of the wrapper classes must be re-done every time the
method interface to the input class changes.<p>

The JDK utility rmic must be applied to the RMI server class after its
construction. The result of that operation will yield the classes
classNameRS_Skel.class and classNameRI_Stub.class<p>

Target and derived class locations for use with RMI ...
<pre>
    Local                                  |       Remote
    -------                                |      --------
    Client Application (using classNameRI) |     className.class
    classNameRI.class                      |     className.class
    classNameRS_stub.class                 |     className_skel.class
</pre>  
@author Chris Anderson (C) UCLA 1998
@version 9/22/98
*/
public class CreateRMIWrappers
{

public CreateRMIWrappers()
{}
/**
Sets the target class. 
@param cName Target class name (full classname - e.g. with package prefix if any).
@exception java.lang.ClassNotFoundException if target class is not found. 
*/
public void setClass(String cName) throws ClassNotFoundException
{
    className = cName;
    classInfo = new ClassInterrogator();
    classInfo.setClass(className);

    localClassName  = null;
    packageName     = null;
    methodString    = null;
    returnString    = null;
    interfaceString = null;
    methodCount     = 0;

    localClassName  = classInfo.getClassName();
    packageName     = classInfo.getPackageName();
    methodString    = classInfo.getMethodNames();
    returnString    = classInfo.getReturnNames();
    interfaceString = classInfo.getInterfaceStrings();

    methodCount      = classInfo.getMethodCount();

    int i;
    parameterStrings = new String [methodCount][];
    argumentStrings  = new String [methodCount][];
    exceptionStrings = new String [methodCount][];

    for(i = 0; i < methodCount; i++)
    {
    parameterStrings[i] = classInfo.getParameterNames(i);
    argumentStrings[i]  = classInfo.getArgumentNames(i);
    exceptionStrings[i] = classInfo.getExceptionNames(i);
    }
}
/**
Returns the name of the target class without the package prefix.
*/
public String getLocalClassName()
{
    return classInfo.getClassName();
}
/**
Creates the RMI server class code. 
@return a String array, each element being a line of the class code.
*/
public String [] createRMIserver()
{
  int i; int j;
  StringBuffer sb;
  Vector cLines = new Vector();
//
// package declaration
//
  if(packageName != null)
  cLines.addElement("package" + " " + packageName + ";");
  
//
// code header
//
  cLines.addElement("//");
  cLines.addElement("//#################################################################");
  cLines.addElement("//      Class " + classInfo.getClassName() +"RS  : " + classInfo.getClassName() + " (R)MI (S)server ");
  cLines.addElement("//#################################################################");
  cLines.addElement("//");
  Date JobDate;
  JobDate = new Date();
  String dateOut =
  DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.SHORT).format(JobDate);
  cLines.addElement("// Construction Date : " + dateOut);
  cLines.addElement("// Auto-generated by CreateRMIWrappers (Chris Anderson (C) UCLA 1998)");
  cLines.addElement("//");
  cLines.addElement("import java.rmi.*;");
  cLines.addElement("import java.rmi.server.*;");
  cLines.addElement(" ");
//
// class declaration
//
  sb = new StringBuffer();
  sb.append("public class " + localClassName + "RS"  + " extends UnicastRemoteObject ");
  cLines.addElement(sb.toString());
  cLines.addElement("implements " + localClassName + "RI");
  cLines.addElement("{");

  cLines.addElement(" ");
  cLines.addElement("private " + className + "  cInstance;");
  cLines.addElement(" ");

  cLines.addElement("public " + localClassName + "RS" + "() throws java.rmi.RemoteException");
  cLines.addElement("{");
  cLines.addElement("    super();");
  cLines.addElement("    cInstance = new " + className + "();");
  cLines.addElement("}");

  for(i = 0; i < methodCount; i++)
  {
     cLines.addElement(" ");

    sb = new StringBuffer();
    sb.append("public "
                             + returnString[i]    + " "
                             + methodString[i]    + "(");

    if(parameterStrings[i].length > 0)
    {sb.append(parameterStrings[i][0] + " " + argumentStrings[i][0]);}
    for(j = 1; j < parameterStrings[i].length; j++)
    {
    sb.append(", " + parameterStrings[i][j] + " " + argumentStrings[i][j]);
    }
    sb.append(") throws java.rmi.RemoteException");

    //
    // Add application specific exceptions thrown
    //
    for(j = 0; j < exceptionStrings[i].length; j++)
    {
    sb.append(", " + exceptionStrings[i][j]);
    }

    cLines.addElement(sb.toString());
    cLines.addElement("{");
    sb = new StringBuffer();

    if(returnString[i].equals("void"))
    {
     sb.append("    cInstance." + methodString[i]+ "(");
    }
    else
    {
    sb.append("    return cInstance." + methodString[i]+ "(");
    }

    if(argumentStrings[i].length > 0)
    {sb.append(argumentStrings[i][0]);}
    for(j = 1; j < argumentStrings[i].length; j++)
    {
    sb.append(", "  + argumentStrings[i][j]);
    }
    sb.append(");");
    cLines.addElement(sb.toString());
    cLines.addElement("}");
  }
//
// Create main() invocation
//
  cLines.addElement("public static void main(String args[])");
  cLines.addElement("{");
  cLines.addElement(" ");
  cLines.addElement("    System.setSecurityManager(new RMISecurityManager());");
  cLines.addElement("   ");
  cLines.addElement("    try {");
  cLines.addElement("    " + localClassName + "RS" + " name = new " + localClassName + "RS();");
  cLines.addElement("    " + "Naming.rebind(\""+ localClassName + "RS\", name);");
  cLines.addElement("    }");
  cLines.addElement("    catch (Exception e) {");
  cLines.addElement("    System.out.println(\"Exception : \" + e.getMessage());");
  cLines.addElement("    e.printStackTrace();");
  cLines.addElement("    }");
  cLines.addElement("}");
//
// class end
//
  cLines.addElement("}");

  String [] returnString = new String[cLines.size()];

  for(i = 0; i < returnString.length; i++)
  {
  returnString[i] = (String)cLines.elementAt(i);
  }

  return returnString;
}
/**
Creates the RMI interface class code. 
@return a String array, each element being a line of the class code.
*/
public String [] createRMIinterface()
{
  int i; int j;
  StringBuffer sb;
  Vector cLines = new Vector();
//
// package declaration
//
  if(packageName != null)
  cLines.addElement("package" + " " + packageName + ";");
//
// code header
//
  cLines.addElement("//");
  cLines.addElement("//#################################################################");
  cLines.addElement("//      Class " + classInfo.getClassName() +"RI  : " + classInfo.getClassName() + " (R)MI (I)nterface ");
  cLines.addElement("//#################################################################");
  cLines.addElement("//");
  Date JobDate;
  JobDate = new Date();
  String dateOut =
  DateFormat.getDateTimeInstance(DateFormat.SHORT,DateFormat.SHORT).format(JobDate);
  cLines.addElement("// Construction Date : " + dateOut);
  cLines.addElement("// Auto-generated by CreateRMIWrappers (Chris Anderson (C) UCLA 1998)");
  cLines.addElement("//");
  cLines.addElement("import java.rmi.*;");
  cLines.addElement("import java.rmi.server.*;");
  cLines.addElement(" ");
//
// class declaration
//
  sb = new StringBuffer();
  sb.append("public interface " + localClassName + "RI"  + " extends java.rmi.Remote");
  if(interfaceString != null)
  {
  for(i = 0; i < interfaceString.length; i++) {sb.append(", " + interfaceString[i]);}
  }
  cLines.addElement(sb.toString());
//
// class definition - all members of the remote interface are public ...
//
  cLines.addElement("{");

  for(i = 0; i < methodCount; i++)
  {
    sb = new StringBuffer();
    sb.append("    "
                             + returnString[i]    + " "
                             + methodString[i]    + "(");
    //
    // add parameters
    //
    if(parameterStrings[i].length > 0)
    {sb.append(parameterStrings[i][0] + " " + argumentStrings[i][0]);}
    for(j = 1; j < parameterStrings[i].length; j++)
    {
    sb.append(", " + parameterStrings[i][j] + " " + argumentStrings[i][j]);
    }
    sb.append(") throws java.rmi.RemoteException");
    //
    // add exceptions
    //
    for(j = 0; j < exceptionStrings[i].length; j++)
    {
    sb.append(", " + exceptionStrings[i][j]);
    }
    sb.append(";");

    cLines.addElement(sb.toString());
    cLines.addElement(" ");
  }
//
// close
//
  cLines.addElement("}");
//
// pack as a string array
//
  String [] returnString = new String[cLines.size()];

  for(i = 0; i < returnString.length; i++)
  {
  returnString[i] = (String)cLines.elementAt(i);
  }

  return returnString;
}

ClassInterrogator classInfo;

String    className;
String    localClassName;
String    packageName;
int       methodCount;
String [] interfaceString;
String [] methodString;
String [] returnString;
String [][] exceptionStrings;
String [][] parameterStrings;
String [][] argumentStrings;
/**
Command line invocation routine. Creates the the interface 
and server classes. Compiles the interface and server classes. 
@param args[] Target class; specified with a full classname 
(e.g. including a package prefix if any).
*/
static public void main(String args[])
{
   int i; int j;
   String inputClass;

   if(args.length == 0) return;
   inputClass = args[0];

   CreateRMIWrappers codeCreator = new CreateRMIWrappers();
   try{codeCreator.setClass(inputClass);}
   catch(ClassNotFoundException e){System.out.println(e.getMessage());}
   String localClassName = codeCreator.getLocalClassName();

   CreateRMIWrappers R = new CreateRMIWrappers();

   try{R.setClass(inputClass);}
   catch(ClassNotFoundException e){System.out.println(e.getMessage());}
   
   System.out.println("Constructing : Interface Class");
   String[] interfaceCode = R.createRMIinterface();
   
   System.out.println("Constructing : Server Class");
   String[] serverCode    = R.createRMIserver();

   File outFileI;
   PrintWriter fileOutputI = null;

   outFileI          = new File(localClassName + "RI.java");
   try{fileOutputI   = new PrintWriter(new FileOutputStream(outFileI),true);}
   catch(IOException e){};

   for(i = 0; i < interfaceCode.length; i++)
   {
    fileOutputI.println(interfaceCode[i]);
   }

   File outFileS;
   PrintWriter fileOutputS = null;

   outFileS          = new File(localClassName + "RS.java");
   try{fileOutputS   = new PrintWriter(new FileOutputStream(outFileS),true);}
   catch(IOException e){};

   for(i = 0; i < serverCode.length; i++)
   {
    fileOutputS.println(serverCode[i]);
   }
   //
   // Invoke the compiler 
   //
   sun.tools.javac.Main compiler = new sun.tools.javac.Main(System.out, "javac");
        
   System.out.println("Compiling : " + localClassName + "RI.java");
   String [] cArgs1 = {localClassName + "RI.java","-O"};
   compiler.compile(cArgs1);
    
   System.out.println("Compiling : " + localClassName + "RS.java");
   String [] cArgs2 = {localClassName + "RS.java","-O"};
   compiler.compile(cArgs2);

   System.out.println("Done : Hit return to continue ");
   BufferedReader D =
   new BufferedReader(new InputStreamReader(System.in));
   try{D.readLine();}catch(Exception e){}
}
}